"""
Run project.yaml actions locally

This creates and runs jobs in a way that's fairly close to what happens in
production, but with the key difference that rather than specifying a repo URL
and a commit we just supply a workspace directory and code is copied into a
Docker volume directly from there. In the past we've had an issue whereby
broken actions work locally by accident because the right output files happen
to exist anyway even though the action doesn't specify that it depends on them.
To try to avoid this, when copying code into a volume we ignore any files which
match any of the output patterns in the project. We then copy in just the
explicit dependencies of the action.

The job creation logic is also slighty different here (compare the
`create_job_request_and_jobs` function below with the `create_jobs` function in
`jobrunner.create_or_update_jobs`) as things like validating the repo URL don't
apply locally.

Other than that, everything else runs entirely as it would in production. A
temporary database and log directory is created for each run and then thrown
away afterwards.
"""
import argparse
import getpass
import os
import random
import shlex
import shutil
import string
import subprocess
import sys
import tempfile
import textwrap
from datetime import datetime, timedelta
from pathlib import Path

from jobrunner import config, executors, manifest_to_database_migration
from jobrunner.create_or_update_jobs import (
    RUN_ALL_COMMAND,
    JobRequestError,
    NothingToDoError,
    ProjectValidationError,
    assert_new_jobs_created,
    get_new_jobs_to_run,
    insert_into_database,
    parse_and_validate_project_file,
)
from jobrunner.lib import database, docker
from jobrunner.lib.database import find_where
from jobrunner.lib.log_utils import configure_logging
from jobrunner.lib.string_utils import tabulate
from jobrunner.lib.subprocess_utils import subprocess_run
from jobrunner.manage_jobs import METADATA_DIR
from jobrunner.models import Job, JobRequest, State, StatusCode, random_id
from jobrunner.project import UnknownActionError, get_all_actions
from jobrunner.queries import calculate_workspace_state
from jobrunner.reusable_actions import (
    ReusableActionError,
    resolve_reusable_action_references,
)
from jobrunner.run import main as run_main


# First paragraph of docstring
DESCRIPTION = __doc__.partition("\n\n")[0]

# local run logging format
LOCAL_RUN_FORMAT = "{action}{message}"


# Super-crude support for colourised/formatted output inside Github Actions. It
# would be good to support formatted output in the CLI more generally, but we
# should use a decent library for that to handle the various cross-platform
# issues.
class ANSI:
    Reset = "\u001b[0m"
    Bold = "\u001b[1m"
    Grey = "\u001b[38;5;248m"


def add_arguments(parser):
    parser.add_argument("actions", nargs="*", help="Name of project action to run")
    parser.add_argument(
        "-f",
        "--force-run-dependencies",
        help="Force the dependencies of the action to run, whether or not their outputs exist",
        action="store_true",
    )
    parser.add_argument(
        "--project-dir",
        help="Project directory (default: current directory)",
        default=".",
    )
    # This only really intended for use in CI. When running locally users
    # generally want to know about failures as soon as they happen whereas in
    # CI you want to run as many actions as possible.
    parser.add_argument(
        "--continue-on-error",
        help="Don't stop on first failed action",
        action="store_true",
    )
    # This particularly useful in CI.
    parser.add_argument(
        "--timestamps",
        help="Include timestamps in output",
        action="store_true",
    )
    parser.add_argument(
        "--debug",
        help="Leave docker containers and volumes in place for debugging",
        action="store_true",
    )
    parser.add_argument(
        "--format-output-for-github",
        help=(
            "Produce output in a format suitable for display inside a "
            "Github Actions Workflow"
        ),
        action="store_true",
    )
    return parser


def main(
    project_dir,
    actions,
    force_run_dependencies=False,
    continue_on_error=False,
    debug=False,
    timestamps=False,
    format_output_for_github=False,
):
    if not docker_preflight_check():
        return False

    project_dir = Path(project_dir).resolve()
    temp_dir = Path(tempfile.mkdtemp(prefix="opensafely_"))
    if not debug:
        # Generate unique docker label to use for all volumes and containers we
        # create during this run in order to make cleanup easy. We're using a
        # random string to prevent mutliple parallel runs from interferring by
        # trying to clean up each others jobs.
        docker_label = "job-runner-local-{}".format(
            "".join(random.choices(string.ascii_uppercase, k=8))
        )
    else:
        # In debug mode (where we don't automatically delete containers and
        # volumes) we use a consistent label to make manual clean up easier
        docker_label = "job-runner-debug"

    log_format = LOCAL_RUN_FORMAT
    if timestamps:
        log_format = "{asctime} " + LOCAL_RUN_FORMAT

    try:
        success_flag = create_and_run_jobs(
            project_dir,
            actions,
            force_run_dependencies=force_run_dependencies,
            continue_on_error=continue_on_error,
            temp_dir=temp_dir,
            docker_label=docker_label,
            clean_up_docker_objects=(not debug),
            log_format=log_format,
            format_output_for_github=format_output_for_github,
        )
    except KeyboardInterrupt:
        print("\nKilled by user")
        print("Cleaning up Docker containers and volumes ...")
        success_flag = False
    finally:
        if not debug:
            delete_docker_entities("container", docker_label, ignore_errors=True)
            delete_docker_entities("volume", docker_label, ignore_errors=True)
            shutil.rmtree(temp_dir, ignore_errors=True)
        else:
            containers = find_docker_entities("container", docker_label)
            volumes = find_docker_entities("volume", docker_label)
            print(f"\n{'-' * 48}")
            print("\nRunning in --debug mode so not cleaning up. To clean up run:\n")
            for container in containers:
                print(f"  docker container rm --force {container}")
            for volume in volumes:
                print(f"  docker volume rm --force {volume}")
            print(f"  rm -rf {temp_dir}")
    return success_flag


def create_and_run_jobs(
    project_dir,
    actions,
    force_run_dependencies,
    continue_on_error,
    temp_dir,
    docker_label,
    clean_up_docker_objects=True,
    log_format=LOCAL_RUN_FORMAT,
    format_output_for_github=False,
):
    # Fiddle with the configuration to suit what we need for running local jobs
    docker.LABEL = docker_label
    # It's more helpful in this context to have things consistent
    config.RANDOMISE_JOB_ORDER = False
    config.HIGH_PRIVACY_WORKSPACES_DIR = project_dir.parent
    config.DATABASE_FILE = project_dir / "metadata" / "db.sqlite"
    config.TMP_DIR = temp_dir
    config.JOB_LOG_DIR = temp_dir / "logs"
    config.BACKEND = "expectations"
    config.USING_DUMMY_DATA_BACKEND = True
    config.CLEAN_UP_DOCKER_OBJECTS = clean_up_docker_objects

    # We want to fetch any reusable actions code directly from Github so as to
    # avoid pushing unnecessary traffic through the proxy
    config.GIT_PROXY_DOMAIN = "github.com"
    # Rather than using the throwaway `temp_dir` to store git repos in we use a
    # consistent directory within the system tempdir. This means we don't have
    # to keep refetching commits and also avoids the complexity of deleting
    # git's read-only directories on Windows. We use the current username as a
    # crude means of scoping the directory to the user in order to avoid
    # potential permissions issues if multiple users share the same directory.
    config.GIT_REPO_DIR = Path(tempfile.gettempdir()).joinpath(
        f"opensafely_{getuser()}"
    )

    # None of the below should be used when running locally
    config.WORKDIR = None
    config.HIGH_PRIVACY_STORAGE_BASE = None
    config.MEDIUM_PRIVACY_STORAGE_BASE = None
    config.MEDIUM_PRIVACY_WORKSPACES_DIR = None

    configure_logging(
        fmt=log_format,
        # All the other output we produce goes to stdout and it's a bit
        # confusing if the log messages end up on a separate stream
        stream=sys.stdout,
        # Filter out log messages in the local run context
        extra_filter=filter_log_messages,
    )

    # This is a temporary migration step to avoid unnecessarily re-running actions as we migrate away from the manifest.
    manifest_to_database_migration.migrate_one(
        project_dir, write_medium_privacy_manifest=False, batch_size=1000, log=False
    )

    # Any jobs that are running or pending must be left over from a previous run that was aborted either by an
    # unexpected and unhandled exception or by the researcher abruptly terminating the process. We can't reasonably
    # recover them (and the researcher may not want to -- maybe that's why they terminated), so we mark them as
    # cancelled. This causes the rest of the system to effectively ignore them.
    #
    # We do this here at the beginning rather than trying to catch these cases when the process exits because the
    # latter couldn't ever completely guarantee to catch every possible termination case correctly.
    database.update_where(
        Job,
        {"cancelled": True, "state": State.FAILED},
        state__in=[State.RUNNING, State.PENDING],
    )

    try:
        job_request, jobs = create_job_request_and_jobs(
            project_dir, actions, force_run_dependencies
        )
    except NothingToDoError:
        print("=> All actions already completed successfully")
        print("   Use -f option to force everything to re-run")
        return True
    except (ProjectValidationError, ReusableActionError, JobRequestError) as e:
        print(f"=> {type(e).__name__}")
        print(textwrap.indent(str(e), "   "))
        if hasattr(e, "valid_actions"):
            print("\n   Valid action names are:")
            for action in e.valid_actions:
                if action != RUN_ALL_COMMAND:
                    print(f"     {action}")
                else:
                    print(f"     {action} (runs all actions in project)")
        return False

    docker_images = get_docker_images(jobs)

    uses_stata = any(
        i.startswith(f"{config.DOCKER_REGISTRY}/stata-mp:") for i in docker_images
    )
    if uses_stata and config.STATA_LICENSE is None:
        config.STATA_LICENSE = get_stata_license()
        if config.STATA_LICENSE is None:
            print(
                "The docker image 'stata-mp' requires a license to function.\n"
                "\n"
                "If you are a member of OpenSAFELY we should have been able to fetch\n"
                "the license automatically, so something has gone wrong. Please open\n"
                "a new discussion here so we can help:\n"
                "  https://github.com/opensafely/documentation/discussions\n"
                "\n"
                "If you are not a member of OpenSAFELY you will have to provide your\n"
                "own license. See the dicussion here for pointers:\n"
                " https://github.com/opensafely/documentation/discussions/299"
            )
            return False

    for image in docker_images:
        if not docker.image_exists_locally(image):
            print(f"Fetching missing docker image: docker pull {image}")
            try:
                # We want to be chatty when running in the console so users can
                # see progress and quiet in CI so we don't spam the logs with
                # layer download noise
                docker.pull(image, quiet=not sys.stdout.isatty())
            except docker.DockerPullError as e:
                print("Failed with error:")
                print(e)
                return False

    action_names = [job.action for job in jobs]
    print(f"\nRunning actions: {', '.join(action_names)}\n")

    # Wrap all the log output inside an expandable block when running inside
    # Github Actions
    if format_output_for_github:
        print(f"::group::Job Runner Logs {ANSI.Grey}(click to view){ANSI.Reset}")

    # Run everything
    exit_condition = (
        no_jobs_remaining if continue_on_error else job_failed_or_none_remaining
    )
    try:
        run_main(exit_callback=exit_condition)
    except KeyboardInterrupt:
        pass
    finally:
        if format_output_for_github:
            print("::endgroup::")

    final_jobs = find_where(
        Job, state__in=[State.FAILED, State.SUCCEEDED], job_request_id=job_request.id
    )
    # Always show failed jobs last, otherwise show in order run
    final_jobs.sort(
        key=lambda job: (
            1 if job.state == State.FAILED else 0,
            job.started_at or 0,
        )
    )

    # Pretty print details of each action
    print()
    if not final_jobs:
        print("=> No jobs completed")
    for job in final_jobs:
        log_file = f"{METADATA_DIR}/{job.action}.log"
        # If a job fails we don't want to clutter the output with its failed
        # dependants.
        if (
            job.state == State.FAILED
            and job.status_code == StatusCode.DEPENDENCY_FAILED
        ):
            continue
        if format_output_for_github:
            print(f"{ANSI.Bold}=> {job.action}{ANSI.Reset}")
        else:
            print(f"=> {job.action}")
        print(textwrap.indent(job.status_message, "   "))
        # Where a job failed because expected outputs weren't found we show a
        # list of other outputs which were generated
        if job.unmatched_outputs:
            print(
                "\n   Did you mean to match one of these files instead?\n    - ", end=""
            )
            print("\n    - ".join(job.unmatched_outputs))
        print()
        # Output the entire log file inside an expandable block when running
        # inside Github Actions
        if format_output_for_github:
            print(
                f"::group:: log file: {log_file} {ANSI.Grey}(click to view){ANSI.Reset}"
            )
            long_grey_line = ANSI.Grey + ("\u2015" * 80) + ANSI.Reset
            print(long_grey_line)
            print((project_dir / log_file).read_text())
            print(long_grey_line)
            print("::endgroup::")
        else:
            print(f"   log file: {log_file}")
        # Display matched outputs
        print("   outputs:")
        outputs = sorted(job.outputs.items()) if job.outputs else []
        print(tabulate(outputs, separator="  - ", indent=5, empty="(no outputs)"))
        # If a job exited with an error code then try to display the end of the
        # log output in case that makes the problem immediately obvious
        if job.status_code == StatusCode.NONZERO_EXIT:
            logs, truncated = get_log_file_snippet(project_dir / log_file, max_lines=32)
            if logs:
                print(f"\n   logs{' (truncated)' if truncated else ''}:\n")
                print(textwrap.indent(logs, "     "))
        print()

    success_flag = all(job.state == State.SUCCEEDED for job in final_jobs)
    return success_flag


def create_job_request_and_jobs(project_dir, actions, force_run_dependencies):
    job_request = JobRequest(
        id=random_id(),
        repo_url=str(project_dir),
        commit=None,
        requested_actions=actions,
        cancelled_actions=[],
        workspace=project_dir.name,
        database_name="dummy",
        force_run_dependencies=force_run_dependencies,
        # The default behaviour of refusing to run if a dependency has failed
        # makes for an awkward workflow when iterating in development
        force_run_failed=True,
        branch="",
        original={"created_by": getuser()},
    )

    project_file_path = project_dir / "project.yaml"
    if not project_file_path.exists():
        raise ProjectValidationError(f"No project.yaml file found in {project_dir}")
    # NOTE: Similar but non-identical logic is implemented for running jobs in
    # production in `jobrunner.create_or_update_jobs.create_jobs`. If you make
    # changes below then consider what, if any, the appropriate corresponding
    # changes might be for production jobs.
    project = parse_and_validate_project_file(project_file_path.read_bytes())
    latest_jobs = calculate_workspace_state(job_request.workspace)

    # On the server out-of-band deletion of an existing output is considered an error, so we ignore that case when
    # scheduling and allow jobs with missing dependencies to fail loudly when they are actually run. However for local
    # running we should allow researchers to delete outputs on disk and automatically rerun the actions that create
    # if they are needed. So here we check whether any files are missing for completed actions and, if so, treat them
    # as though they had not been run -- this will automatically trigger a rerun.
    latest_jobs_with_files_present = [
        job for job in latest_jobs if all_output_files_present(project_dir, job)
    ]

    try:
        if not actions:
            raise UnknownActionError("At least one action must be supplied")
        new_jobs = get_new_jobs_to_run(
            job_request, project, latest_jobs_with_files_present
        )
    except UnknownActionError as e:
        # Annotate the exception with a list of valid action names so we can
        # show them to the user
        e.valid_actions = [RUN_ALL_COMMAND] + get_all_actions(project)
        raise e
    assert_new_jobs_created(new_jobs, latest_jobs_with_files_present)
    resolve_reusable_action_references(new_jobs)
    insert_into_database(job_request, new_jobs)
    return job_request, new_jobs


def all_output_files_present(project_dir, job):
    return all(project_dir.joinpath(f).exists() for f in job.output_files)


def no_jobs_remaining(active_jobs):
    return len(active_jobs) == 0


def job_failed_or_none_remaining(active_jobs):
    if any(job.state == State.FAILED for job in active_jobs):
        return True
    return len(active_jobs) == 0


def filter_log_messages(record):
    """
    Not all log messages are useful in the local run context so to avoid noise
    and make things clearer for the user we filter them out here
    """
    # None of these status messages are particularly useful in local run
    # mode, and they can generate a lot of clutter in large dependency
    # trees
    if getattr(record, "status_code", None) in {
        StatusCode.WAITING_ON_DEPENDENCIES,
        StatusCode.DEPENDENCY_FAILED,
        StatusCode.WAITING_ON_WORKERS,
    }:
        return False

    # We sometimes log caught exceptions for debugging purposes in production,
    # but we don't want to show these to the user when running locally
    if getattr(record, "exc_info", None):
        return False

    # Executor state logging is pretty verbose and unlikely to be useful for local running
    if record.name == executors.logging.LOGGER_NAME:
        return False

    return True


# Copied from test/conftest.py to avoid a more complex dependency tree
def delete_docker_entities(entity, label, ignore_errors=False):
    ls_args = [
        "docker",
        entity,
        "ls",
        "--all" if entity == "container" else None,
        "--filter",
        f"label={label}",
        "--quiet",
    ]
    ls_args = list(filter(None, ls_args))
    response = subprocess_run(
        ls_args, capture_output=True, encoding="ascii", check=not ignore_errors
    )
    ids = response.stdout.split()
    if ids and response.returncode == 0:
        rm_args = ["docker", entity, "rm", "--force"] + ids
        subprocess_run(rm_args, capture_output=True, check=not ignore_errors)


def find_docker_entities(entity, label):
    """
    Return list of names of all docker entities (of specified type) matching
    `label`
    """
    response = subprocess_run(
        [
            "docker",
            entity,
            "ls",
            *(["--all"] if entity == "container" else []),
            "--filter",
            f"label={label}",
            "--format",
            "{{ .Names }}" if entity == "container" else "{{ .Name }}",
            "--quiet",
        ],
        capture_output=True,
        encoding="ascii",
    )
    return response.stdout.split()


def get_docker_images(jobs):
    docker_images = {shlex.split(job.run_command)[0] for job in jobs}
    full_docker_images = {
        f"{config.DOCKER_REGISTRY}/{image}" for image in docker_images
    }
    # We always need this image to work with volumes
    full_docker_images.add(docker.MANAGEMENT_CONTAINER_IMAGE)
    return full_docker_images


def get_log_file_snippet(log_file, max_lines):
    try:
        contents = Path(log_file).read_text()
    except Exception:
        contents = ""
    # As docker logs are timestamp-prefixed the first blank line marks the end
    # of the docker logs and the start of our "trailer"
    docker_logs = contents.partition("\n\n")[0]
    # Strip off timestamp
    log_lines = [line[31:] for line in docker_logs.splitlines()]
    if len(log_lines) > max_lines:
        log_lines = log_lines[-max_lines:]
        truncated = True
    else:
        truncated = False
    return "\n".join(log_lines).strip(), truncated


def get_stata_license(repo=config.STATA_LICENSE_REPO):
    """Load a stata license from local cache or remote repo."""
    cached = Path(f"{tempfile.gettempdir()}/opensafely-stata.lic")
    license_timeout = timedelta(hours=2)

    def git_clone(repo_url, cwd):
        cmd = ["git", "clone", "--depth=1", repo_url, "repo"]
        # GIT_TERMINAL_PROMPT=0 means it will fail if it requires auth. This
        # allows us to retry with an ssh url on linux/mac, as they would
        # generally prompt given an https url.
        result = subprocess_run(
            cmd,
            cwd=cwd,
            capture_output=True,
            env=dict(os.environ, GIT_TERMINAL_PROMPT="0"),
        )
        return result.returncode == 0

    fetch = False
    if cached.exists():
        mtime = datetime.fromtimestamp(cached.stat().st_mtime)
        if datetime.utcnow() - mtime > license_timeout:
            fetch = True
    else:
        fetch = True

    if fetch:
        try:
            tmp = tempfile.TemporaryDirectory(suffix="opensafely")
            success = git_clone(repo, tmp.name)
            # http urls usually won't work for linux/mac clients, so try ssh
            if not success and repo.startswith("https://"):
                git_clone(
                    repo.replace("https://", "git+ssh://git@"),
                    tmp.name,
                )
            shutil.copyfile(f"{tmp.name}/repo/stata.lic", cached)
        except Exception:
            pass
        finally:
            tmp.cleanup()

    if cached.exists():
        # if the refresh failed for some reason, update the last time it was
        # used to now to avoid spamming github on every subsequent run
        t = datetime.utcnow().timestamp()
        os.utime(cached, (t, t))
        return cached.read_text()
    else:
        return None


def docker_preflight_check():
    try:
        subprocess_run(["docker", "info"], check=True, capture_output=True)
    except FileNotFoundError:
        print("Could not find application: docker")
        print("\nYou must have Docker installed to run this command, see:")
        print("https://docs.docker.com/get-docker/")
        return False
    except subprocess.CalledProcessError:
        print("There was an error running: docker info")
        print("\nIt looks like you have Docker installed but have not started it.")
        return False
    return True


def getuser():
    # `getpass.getuser()` can fail on Windows under certain circumstances. It's never
    # critical that we know the current username so we don't want to block execution if
    # this happens
    try:
        return getpass.getuser()
    except Exception:
        return "unknown"


def run():
    parser = argparse.ArgumentParser(description=DESCRIPTION)
    parser = add_arguments(parser)
    args = parser.parse_args()
    success = main(**vars(args))
    sys.exit(0 if success else 1)


if __name__ == "__main__":
    run()



import subprocess

import pytest

from jobrunner.lib import docker


@pytest.mark.slow_test
@pytest.mark.needs_docker
def test_basic_volume_interaction(tmp_path, docker_cleanup):
    files = ["test1.txt", "test2.json", "subdir/test3.txt", "subdir/test4.json"]
    for name in files:
        path = tmp_path / name
        path.parent.mkdir(parents=True, exist_ok=True)
        path.touch()
    volume = "jobrunner-volume-test"
    docker.create_volume(volume)
    # Test no error is thrown if volume already exists
    docker.create_volume(volume)
    docker.copy_to_volume(volume, tmp_path, ".")
    matches = docker.glob_volume_files(volume, ["*.txt", "*.json"])
    assert matches == {"*.txt": ["test1.txt"], "*.json": ["test2.json"]}
    matches = docker.glob_volume_files(volume, ["subdir/*"])
    assert matches == {"subdir/*": ["subdir/test3.txt", "subdir/test4.json"]}
    docker.delete_volume(volume)
    # Test no error is thrown if volume is already deleted
    docker.delete_volume(volume)


def test_disk_space_detection(monkeypatch):
    def error(stdout, stderr):
        def run(args, timeout, **kwargs):
            raise subprocess.CalledProcessError(
                returncode=1,
                cmd=args,
                output=stdout,
                stderr=stderr,
            )

        return run

    msg = "Error response from daemon: no space left on device"

    monkeypatch.setattr(docker, "subprocess_run", error(msg, None))
    with pytest.raises(docker.DockerDiskSpaceError):
        docker.docker([])

    monkeypatch.setattr(docker, "subprocess_run", error(None, msg))
    with pytest.raises(docker.DockerDiskSpaceError):
        docker.docker([])

    msg = msg.encode("utf8")

    monkeypatch.setattr(docker, "subprocess_run", error(msg, None))
    with pytest.raises(docker.DockerDiskSpaceError):
        docker.docker([])

    monkeypatch.setattr(docker, "subprocess_run", error(None, msg))
    with pytest.raises(docker.DockerDiskSpaceError):
        docker.docker([])


@pytest.mark.needs_docker
def test_copy_to_and_from_volume(tmp_path, docker_cleanup):
    volume = __name__
    src = tmp_path / "src.txt"
    dst = tmp_path / "dst.txt"
    src.write_text("I exist")
    docker.create_volume(volume)
    docker.copy_to_volume(volume, src, "src.txt")
    docker.copy_from_volume(volume, "src.txt", dst)
    assert dst.read_text() == "I exist"
    assert len(list(tmp_path.glob("dst.txt*.tmp"))) == 0


@pytest.mark.needs_docker
def test_copy_to_volume_dereference_symlinks(tmp_path, docker_cleanup):
    volume = __name__
    target = tmp_path / "target.txt"
    target.write_text("target")
    link = tmp_path / "link.txt"
    link.symlink_to(target)
    dst = tmp_path / "dst.txt"

    docker.create_volume(volume)
    docker.copy_to_volume(volume, link, "src.txt")
    docker.copy_from_volume(volume, "src.txt", dst)
    assert not dst.is_symlink()
    assert dst.read_text() == "target"


@pytest.mark.needs_docker
def test_copy_from_volume_error(tmp_path, docker_cleanup, monkeypatch):
    volume = __name__
    src = tmp_path / "src.txt"
    dst = tmp_path / "dst.txt"

    src.write_text("I exist")
    docker.create_volume(volume)
    docker.copy_to_volume(volume, src, "src.txt")

    def docker_error(*args, **kwards):
        raise Exception("oh noes")

    monkeypatch.setattr(docker, "docker", docker_error)

    with pytest.raises(Exception):
        docker.copy_from_volume(volume, "src.txt", dst)

    assert not dst.exists()
    assert len(list(tmp_path.glob("dst.txt*.tmp"))) == 0



from collections import deque

import hypothesis.strategies as st
import pytest
from hypothesis import assume, settings
from hypothesis.stateful import Bundle, RuleBasedStateMachine, invariant, rule

from jobrunner.lib.lru_dict import LRUDict


@settings(max_examples=500)
class LRUDictValidation(RuleBasedStateMachine):
    capacity = 5
    keys = st.integers(min_value=1)
    values = st.integers(min_value=1)
    existing_items = Bundle("existing_items")

    def __init__(self):
        super().__init__()
        self.values = dict()
        self.recent_keys = deque(maxlen=self.capacity)
        self.lru_dict = LRUDict(self.capacity)

    @rule(key=keys, value=values, target=existing_items)
    def insert_item_for_the_first_time(self, key, value):
        self._insert(key, value)
        return key, value

    @rule(item=existing_items)
    def reinsert_an_existing_item_unchanged(self, item):
        key, value = item
        self._insert(key, value)

    @rule(item=existing_items, value=values, target=existing_items)
    def overwrite_an_existing_item(self, item, value):
        key, _ = item
        self._insert(key, value)
        return key, value

    @rule(item=existing_items)
    def get_item_thats_present(self, item):
        key, _ = item
        assume(key in self.recent_keys)
        self.lru_dict.get(key)
        self._record_access(key)

    @rule(item=existing_items)
    def get_item_thats_been_evicted(self, item):
        key, _ = item
        assume(key not in self.recent_keys)
        self.lru_dict.get(key)

    @rule(key=keys)
    def get_item_thats_never_been_inserted(self, key):
        assume(key not in self.values)
        self.lru_dict.get(key)

    @invariant()
    def is_the_right_length(self):
        assert len(self.lru_dict) == len(self.recent_keys)

    @invariant()
    def has_a_bounded_length(self):
        assert len(self.lru_dict) <= self.capacity

    @invariant()
    def contains_the_right_items(self):
        assert set(self.lru_dict.keys()) == set(self.recent_keys)

    @invariant()
    def has_the_right_values(self):
        # Note that this operation accesses the values in the LRUDict, which has the side-effect of updating their
        # recency; at first blush this is undesirable because we don't want our invariant checks to modify the data.
        # However in this case it's okay because we touch every item in exactly their recency order (oldest to newest)
        # because were iterating through the recent keys in order; so the entire operation leaves the recency exactly
        # as it was.
        #
        # We have to jigger around getting a static list of the keys because the get implementation for this dict can
        # modify the data, which is illegal during iteration.
        keys = list(self.lru_dict.keys())
        for key in keys:
            assert self.lru_dict[key] == self.values[key]

    def _insert(self, key, value):
        self.lru_dict[key] = value
        self.values[key] = value
        self._record_access(key)

    def _record_access(self, key):
        if key in self.recent_keys:
            self.recent_keys.remove(key)
        self.recent_keys.append(key)


LRUDictTest = pytest.mark.slow_test(LRUDictValidation.TestCase)



"""
This module provides a single public entry point `create_or_update_jobs`.

It handles all logic connected with creating or updating Jobs in response to
JobRequests. This includes fetching the code with git, validating the project
and doing the necessary dependency resolution.
"""
import logging
import re
import time

from jobrunner import config
from jobrunner.lib.database import exists_where, insert, transaction, update_where
from jobrunner.lib.git import GitError, GitFileNotFoundError, read_file_from_repo
from jobrunner.lib.github_validators import (
    GithubValidationError,
    validate_branch_and_commit,
    validate_repo_url,
)
from jobrunner.models import Job, SavedJobRequest, State
from jobrunner.project import (
    RUN_ALL_COMMAND,
    ProjectValidationError,
    get_action_specification,
    get_all_actions,
    parse_and_validate_project_file,
)
from jobrunner.queries import calculate_workspace_state
from jobrunner.reusable_actions import (
    ReusableActionError,
    resolve_reusable_action_references,
)


log = logging.getLogger(__name__)


class JobRequestError(Exception):
    pass


class NothingToDoError(JobRequestError):
    pass


def create_or_update_jobs(job_request):
    """
    Create or update Jobs in response to a JobRequest

    Note that where there is an error with the JobRequest it will create a
    single, failed job with the error details rather than raising an exception.
    This allows the error to be synced back to the job-server where it can be
    displayed to the user.
    """
    if not related_jobs_exist(job_request):
        try:
            log.info(f"Handling new JobRequest:\n{job_request}")
            new_job_count = create_jobs(job_request)
            log.info(f"Created {new_job_count} new jobs")
        except (
            GitError,
            GithubValidationError,
            ProjectValidationError,
            ReusableActionError,
            JobRequestError,
        ) as e:
            log.info(f"JobRequest failed:\n{e}")
            create_failed_job(job_request, e)
        except Exception:
            log.exception("Uncaught error while creating jobs")
            create_failed_job(job_request, JobRequestError("Internal error"))
    else:
        if job_request.cancelled_actions:
            log.debug("Cancelling actions: %s", job_request.cancelled_actions)
            set_cancelled_flag_for_actions(
                job_request.id, job_request.cancelled_actions
            )
        else:
            log.debug("Ignoring already processed JobRequest")


def create_jobs(job_request):
    # NOTE: Similar but non-identical logic is implemented for running jobs
    # locally in `jobrunner.cli.local_run.create_job_request_and_jobs`. If you
    # make changes below then consider what the appropriate corresponding
    # changes are for locally run jobs.
    validate_job_request(job_request)
    project_file = get_project_file(job_request)
    project = parse_and_validate_project_file(project_file)
    latest_jobs = get_latest_jobs_for_actions_in_project(job_request.workspace, project)
    new_jobs = get_new_jobs_to_run(job_request, project, latest_jobs)
    assert_new_jobs_created(new_jobs, latest_jobs)
    resolve_reusable_action_references(new_jobs)
    # There is a delay between getting the current jobs (which we fetch from
    # the database and the disk) and inserting our new jobs below. This means
    # the state of the world may have changed in the meantime. Why is this OK?
    #
    # Because we're single threaded and because this function is the only place
    # jobs are created, we can guarantee that no *new* jobs were created. So
    # the only state change that's possible is that some active jobs might have
    # completed. That's unproblematic: any new jobs which are waiting on these
    # now-already-completed jobs will see they have completed the first time
    # they check and then proceed as normal.
    #
    # (It is also possible that someone could delete files off disk that are
    # needed by a particular job, but there's not much we can do about that
    # other than fail gracefully when trying to start the job.)
    insert_into_database(job_request, new_jobs)
    return len(new_jobs)


def validate_job_request(job_request):
    if config.ALLOWED_GITHUB_ORGS:
        validate_repo_url(job_request.repo_url, config.ALLOWED_GITHUB_ORGS)
    if not job_request.requested_actions:
        raise JobRequestError("At least one action must be supplied")
    if not job_request.workspace:
        raise JobRequestError("Workspace name cannot be blank")
    if re.search(r"[^a-zA-Z0-9_\-]", job_request.workspace):
        raise JobRequestError(
            "Invalid workspace name (allowed are alphanumeric, dash and underscore)"
        )
    if not config.USING_DUMMY_DATA_BACKEND:
        database_name = job_request.database_name
        valid_names = config.DATABASE_URLS.keys()

        if database_name not in valid_names:
            raise JobRequestError(
                f"Invalid database name '{database_name}', allowed are: "
                + ", ".join(valid_names)
            )

        if not config.DATABASE_URLS[database_name]:
            raise JobRequestError(
                f"Database name '{database_name}' is not currently defined "
                f"for backend '{config.BACKEND}'"
            )
    # If we're not restricting to specific Github organisations then there's no
    # point in checking the provenance of the supplied commit
    if config.ALLOWED_GITHUB_ORGS:
        # As this involves talking to the remote git server we only do it at
        # the end once all other checks have passed
        validate_branch_and_commit(
            job_request.repo_url, job_request.commit, job_request.branch
        )


def get_project_file(job_request):
    try:
        return read_file_from_repo(
            job_request.repo_url, job_request.commit, "project.yaml"
        )
    except GitFileNotFoundError:
        raise JobRequestError(f"No project.yaml file found in {job_request.repo_url}")


def get_latest_jobs_for_actions_in_project(workspace, project):
    return [
        job
        for job in calculate_workspace_state(workspace)
        if job.action in get_all_actions(project)
    ]


def get_new_jobs_to_run(job_request, project, current_jobs):
    """
    Returns a list of new jobs to run in response to the supplied JobReqeust

    Args:
        job_request: JobRequest instance
        project: dict representing the parsed project file from the JobRequest
        current_jobs: list containing the most recent Job for each action in
            the workspace
    """
    # Build a dict mapping action names to job instances
    jobs_by_action = {job.action: job for job in current_jobs}
    # Add new jobs to it by recursing through the dependency tree
    for action in get_actions_to_run(job_request, project):
        recursively_build_jobs(jobs_by_action, job_request, project, action)

    # Pick out the new jobs we've added and return them
    return [job for job in jobs_by_action.values() if job not in current_jobs]


def get_actions_to_run(job_request, project):
    # Handle the special `run_all` action
    if RUN_ALL_COMMAND in job_request.requested_actions:
        return get_all_actions(project)
    else:
        return job_request.requested_actions


def recursively_build_jobs(jobs_by_action, job_request, project, action):
    """
    Recursively populate the `jobs_by_action` dict with jobs

    Args:
        jobs_by_action: A dict mapping action ID strings to Job instances
        job_request: An instance of JobRequest representing the job request.
        project: A dict representing the project.
        action: The string ID of the action to be added as a job.
    """
    existing_job = jobs_by_action.get(action)
    if existing_job and not job_should_be_rerun(job_request, existing_job):
        return

    action_spec = get_action_specification(project, action)

    # Walk over the dependencies of this action, creating any necessary jobs,
    # and ensure that this job waits for its dependencies to finish before it
    # starts
    wait_for_job_ids = []
    for required_action in action_spec.needs:
        recursively_build_jobs(jobs_by_action, job_request, project, required_action)
        required_job = jobs_by_action[required_action]
        if required_job.state in [State.PENDING, State.RUNNING]:
            wait_for_job_ids.append(required_job.id)

    job = Job(
        job_request_id=job_request.id,
        state=State.PENDING,
        repo_url=job_request.repo_url,
        commit=job_request.commit,
        workspace=job_request.workspace,
        database_name=job_request.database_name,
        action=action,
        wait_for_job_ids=wait_for_job_ids,
        requires_outputs_from=action_spec.needs,
        run_command=action_spec.run,
        output_spec=action_spec.outputs,
        created_at=int(time.time()),
        updated_at=int(time.time()),
    )

    # Add it to the dictionary of scheduled jobs
    jobs_by_action[action] = job


def job_should_be_rerun(job_request, job):
    """
    Do we need to run the action referenced by this job again?
    """
    # Already running or about to run so don't start a new one
    if job.state in [State.PENDING, State.RUNNING]:
        return False
    # Explicitly requested actions always get re-run
    if job.action in job_request.requested_actions:
        return True
    # If it's not an explicilty requested action then it's a dependency, and if
    # we're forcing all dependencies to run then we need to run this one
    if job_request.force_run_dependencies:
        return True

    # Otherwise if it succeeded last time there's no need to run again
    if job.state == State.SUCCEEDED:
        return False
    # If it failed last time ...
    elif job.state == State.FAILED:
        # ... and we're forcing failed jobs to re-run then re-run it
        if job_request.force_run_failed:
            return True
        # Otherwise it's an error condition
        raise JobRequestError(
            f"{job.action} failed on a previous run and must be re-run"
        )
    else:
        raise ValueError(f"Invalid state: {job}")


def assert_new_jobs_created(new_jobs, current_jobs):
    if new_jobs:
        return

    pending = [job.action for job in current_jobs if job.state == State.PENDING]
    running = [job.action for job in current_jobs if job.state == State.RUNNING]
    succeeded = [job.action for job in current_jobs if job.state == State.SUCCEEDED]
    failed = [job.action for job in current_jobs if job.state == State.FAILED]

    # There are two legitimate reasons we can end up with no new jobs to run...
    if len(succeeded) == len(current_jobs):
        #  ...one is that the "run all" action was requested but everything has already run successfully
        log.info(
            f"run_all requested, but all jobs are already successful: {', '.join(succeeded)}"
        )
        raise NothingToDoError()

    if len(pending) + len(running) == len(current_jobs):
        # ...the other is that every requested action is already running or pending, this is considered a user error
        statuses = ", ".join(f"{job.action}({job.state.name})" for job in current_jobs)
        log.info(f"All requested actions were already scheduled to run: {statuses}")
        raise JobRequestError("All requested actions were already scheduled to run")

    # But if we get here then we've somehow failed to schedule new jobs despite the fact that some of the actions we
    # depend on have failed, which is a bug.
    raise Exception(
        f"Unexpected failed jobs in dependency graph after scheduling: {', '.join(failed)}"
    )


def create_failed_job(job_request, exception):
    """
    Sometimes we want to say to the job-server (and the user): your JobRequest
    was broken so we weren't able to create any jobs for it. But the only way
    for the job-runner to communicate back to the job-server is by creating a
    job. So this function creates a single job with the special action name
    "__error__", which starts in the FAILED state and whose status_message
    contains the error we wish to communicate.

    This is a bit of a hack, but it keeps the sync protocol simple.
    """
    # Special case for the NothingToDoError which we treat as a success
    if isinstance(exception, NothingToDoError):
        state = State.SUCCEEDED
        status_message = "All actions have already run"
        action = job_request.requested_actions[0]
    else:
        state = State.FAILED
        status_message = f"{type(exception).__name__}: {exception}"
        action = "__error__"
    now = int(time.time())
    job = Job(
        job_request_id=job_request.id,
        state=state,
        repo_url=job_request.repo_url,
        commit=job_request.commit,
        workspace=job_request.workspace,
        action=action,
        status_message=status_message,
        created_at=now,
        started_at=now,
        updated_at=now,
        completed_at=now,
    )
    insert_into_database(job_request, [job])


def insert_into_database(job_request, jobs):
    with transaction():
        insert(SavedJobRequest(id=job_request.id, original=job_request.original))
        for job in jobs:
            insert(job)


def related_jobs_exist(job_request):
    return exists_where(Job, job_request_id=job_request.id)


def set_cancelled_flag_for_actions(job_request_id, actions):
    # It's important that we modify the Jobs in-place in the database rather than retrieving, updating and re-writing
    # them. If we did the latter then we would risk dirty writes if the run thread modified a Job while we were
    # working.
    update_where(
        Job,
        {"cancelled": True},
        job_request_id=job_request_id,
        action__in=actions,
    )



"""
Script which polls the job-server endpoint for active JobRequests and POSTs
back any associated Jobs.
"""
import logging
import sys
import time

import requests

from jobrunner import config
from jobrunner.create_or_update_jobs import create_or_update_jobs
from jobrunner.lib.database import find_where
from jobrunner.lib.log_utils import configure_logging, set_log_context
from jobrunner.models import Job, JobRequest


session = requests.Session()
log = logging.getLogger(__name__)


class SyncAPIError(Exception):
    pass


def main():
    log.info(
        f"Polling for JobRequests at: "
        f"{config.JOB_SERVER_ENDPOINT.rstrip('/')}/job-requests/"
    )
    while True:
        sync()
        time.sleep(config.POLL_INTERVAL)


def sync():
    response = api_get(
        "job-requests",
        # We're deliberately not paginating here on the assumption that the set
        # of active jobs is always going to be small enough that we can fetch
        # them in a single request and we don't need the extra complexity
        params={"backend": config.BACKEND},
    )
    job_requests = [job_request_from_remote_format(i) for i in response["results"]]

    # Bail early if there's nothing to do
    if not job_requests:
        return

    job_request_ids = [i.id for i in job_requests]
    for job_request in job_requests:
        with set_log_context(job_request=job_request):
            create_or_update_jobs(job_request)
    jobs = find_where(Job, job_request_id__in=job_request_ids)
    jobs_data = [job_to_remote_format(i) for i in jobs]
    log.debug(f"Syncing {len(jobs_data)} jobs back to job-server")

    api_post("jobs", json=jobs_data)


def api_get(*args, **kwargs):
    return api_request("get", *args, **kwargs)


def api_post(*args, **kwargs):
    return api_request("post", *args, **kwargs)


def api_request(method, path, *args, **kwargs):
    url = "{}/{}/".format(config.JOB_SERVER_ENDPOINT.rstrip("/"), path.strip("/"))
    # We could do this just once on import, but it makes changing the config in
    # tests more fiddly
    session.headers = {"Authorization": config.JOB_SERVER_TOKEN}
    response = session.request(method, url, *args, **kwargs)

    log.debug(
        "%s %s %s post_data=%s %s"
        % (
            method.upper(),
            response.status_code,
            url,
            kwargs.get("json", '""'),
            response.text,
        )
    )

    try:
        response.raise_for_status()
    except Exception as e:
        raise SyncAPIError(e) from e

    return response.json()


def job_request_from_remote_format(job_request):
    """
    Convert a JobRequest as received from the job-server into our own internal
    representation
    """
    return JobRequest(
        id=str(job_request["identifier"]),
        repo_url=job_request["workspace"]["repo"],
        commit=job_request["sha"],
        branch=job_request["workspace"]["branch"],
        requested_actions=job_request["requested_actions"],
        cancelled_actions=job_request["cancelled_actions"],
        workspace=job_request["workspace"]["name"],
        database_name=job_request["workspace"]["db"],
        force_run_dependencies=job_request["force_run_dependencies"],
        original=job_request,
    )


def job_to_remote_format(job):
    """
    Convert our internal representation of a Job into whatever format the
    job-server expects
    """
    return {
        "identifier": job.id,
        "job_request_id": job.job_request_id,
        "action": job.action,
        "status": job.state.value,
        "status_code": job.status_code.value if job.status_code else "",
        "status_message": job.status_message or "",
        "created_at": job.created_at_isoformat,
        "updated_at": job.updated_at_isoformat,
        "started_at": job.started_at_isoformat,
        "completed_at": job.completed_at_isoformat,
    }


if __name__ == "__main__":
    configure_logging()

    try:
        main()
    except KeyboardInterrupt:
        sys.exit(0)



import argparse
import json
import logging
import os
import shutil
import time
from datetime import datetime, timedelta
from pathlib import Path

import pytest

from jobrunner import config
from jobrunner.cli import local_run
from jobrunner.lib import database
from jobrunner.lib.subprocess_utils import subprocess_run
from jobrunner.manage_jobs import MANIFEST_FILE, METADATA_DIR
from jobrunner.models import Job, SavedJobRequest, State
from jobrunner.project import get_action_specification, parse_and_validate_project_file


FIXTURE_DIR = Path(__file__).parents[1].resolve() / "fixtures"


@pytest.fixture
def use_api(request, monkeypatch):
    monkeypatch.setattr(config, "EXECUTION_API", request.param)


@pytest.mark.parametrize("use_api", [True, False], indirect=True)
@pytest.mark.parametrize("extraction_tool", ["cohortextractor", "databuilder"])
@pytest.mark.slow_test
@pytest.mark.needs_docker
def test_local_run_success(use_api, extraction_tool, tmp_path, docker_cleanup):
    project_dir = tmp_path / "project"
    shutil.copytree(str(FIXTURE_DIR / "full_project"), project_dir)

    local_run.main(project_dir=project_dir, actions=[f"analyse_data_{extraction_tool}"])

    # FIXME: consolidate these when databuilder supports more columns in dummy data
    if extraction_tool == "cohortextractor":
        paths = [
            "output/input.csv",
            "cohortextractor-counts.txt",
            "metadata/analyse_data_cohortextractor.log",
            "metadata/db.sqlite",
        ]
    else:
        paths = [
            "output/dataset.csv",
            "output/count_by_year.csv",
            "metadata/analyse_data_databuilder.log",
            "metadata/db.sqlite",
        ]

    for path in paths:
        assert (project_dir / path).exists(), path
    assert not (project_dir / "metadata/.logs").exists()


@pytest.mark.slow_test
@pytest.mark.needs_docker
@pytest.mark.skipif(
    not os.environ.get("STATA_LICENSE"), reason="No STATA_LICENSE env var"
)
def test_local_run_stata(tmp_path, monkeypatch, docker_cleanup):
    project_dir = tmp_path / "project"
    shutil.copytree(str(FIXTURE_DIR / "stata_project"), project_dir)
    monkeypatch.setattr("jobrunner.config.STATA_LICENSE", os.environ["STATA_LICENSE"])
    local_run.main(project_dir=project_dir, actions=["stata"])
    env_file = project_dir / "output/env.txt"
    assert "Bennett Institute" in env_file.read_text()


@pytest.mark.slow_test
@pytest.mark.needs_docker
def test_local_run_triggers_a_manifest_migration(tmp_path, docker_cleanup):
    project_dir = tmp_path / "project"
    shutil.copytree(str(FIXTURE_DIR / "full_project"), project_dir)

    # This action doesn't exist in the project.yaml, but the migration doesn't care. We use this instead of an action
    # that does exist so that it's unambiguous that the database record had been created by the migration rather than
    # as a side-effect of running the action we specify.
    manifest = {
        "workspace": "the-workspace",
        "repo": "the-repo-url",
        "actions": {"the-action": {"job_id": "job-id-from-manifest"}},
        "files": {},
    }
    manifest_file = project_dir / METADATA_DIR / MANIFEST_FILE
    manifest_file.parent.mkdir(parents=True)
    manifest_file.write_text(json.dumps(manifest))

    local_run.main(project_dir=project_dir, actions=["generate_cohort"])

    assert database.exists_where(Job, id="job-id-from-manifest")


@pytest.mark.slow_test
@pytest.mark.needs_docker
@pytest.mark.parametrize("extraction_tool", ["cohortextractor", "databuilder"])
def test_local_run_copes_with_detritus_of_earlier_interrupted_run(
    extraction_tool, tmp_path
):
    # This test simulates the case where an earlier run has been interrupted (for example by the user pressing ctrl-c).
    # In particular we put a couple of jobs in unfinished states, which they could never be left in under normal
    # operation. The correct behaviour of the local run, which this tests for, is for such unfinished jobs to be marked
    # as cancelled on the next run.
    project_dir = tmp_path / "project"
    shutil.copytree(str(FIXTURE_DIR / "full_project"), project_dir)
    config.DATABASE_FILE = project_dir / "metadata" / "db.sqlite"

    project = parse_and_validate_project_file(
        (project_dir / "project.yaml").read_bytes()
    )
    database.insert(SavedJobRequest(id="previous-request", original={}))

    def job(job_id, action, state):
        spec = get_action_specification(project, action)
        return Job(
            id=job_id,
            job_request_id="previous-request",
            state=state,
            status_message="",
            repo_url=str(project_dir),
            workspace=project_dir.name,
            database_name="a-database",
            action=action,
            wait_for_job_ids=[],
            requires_outputs_from=spec.needs,
            run_command=spec.run,
            output_spec=spec.outputs,
            created_at=int(time.time()),
            updated_at=int(time.time()),
            outputs={},
        )

    # FIXME: consolidate these when databuilder supports more columns in dummy data
    if extraction_tool == "cohortextractor":
        actions = ["generate_cohort", "prepare_data_m_cohortextractor"]
    else:
        actions = ["generate_dataset", "analyse_data_databuilder"]

    database.insert(job(job_id="123", action=actions[0], state=State.RUNNING))
    database.insert(job(job_id="456", action=actions[1], state=State.PENDING))
    assert local_run.main(project_dir=project_dir, actions=[actions[1]])

    assert database.find_one(Job, id="123").cancelled
    assert database.find_one(Job, id="123").state == State.FAILED
    assert database.find_one(Job, id="456").cancelled
    assert database.find_one(Job, id="456").state == State.FAILED


@pytest.fixture
def systmpdir(monkeypatch, tmp_path):
    """Set the system tempdir to tmp_path for this test, for isolation."""
    monkeypatch.setattr("tempfile.tempdir", str(tmp_path))


@pytest.fixture
def license_repo(tmp_path):
    # create a repo to clone the license from
    repo = tmp_path / "test-repo"
    repo.mkdir()
    license = repo / "stata.lic"  # noqa: A001
    license.write_text("repo-license")
    git = ["git", "-c", "user.name=test", "-c", "user.email=test@example.com"]
    env = {"GIT_CONFIG_GLOBAL": "/dev/null"}
    repo_path = str(repo)
    subprocess_run(git + ["init"], cwd=repo_path, env=env)
    subprocess_run(git + ["add", "stata.lic"], cwd=repo_path, env=env)
    subprocess_run(
        git + ["commit", "--no-gpg-sign", "-m", "test"], cwd=repo_path, env=env
    )
    return repo_path


def test_get_stata_license_cache_recent(systmpdir, monkeypatch, tmp_path):
    def fail(*a, **kwargs):
        assert False, "should not have been called"

    monkeypatch.setattr("jobrunner.lib.subprocess_utils.subprocess_run", fail)
    cache = tmp_path / "opensafely-stata.lic"
    cache.write_text("cached-license")
    assert local_run.get_stata_license() == "cached-license"


def test_get_stata_license_cache_expired(systmpdir, tmp_path, license_repo):
    cache = tmp_path / "opensafely-stata.lic"
    cache.write_text("cached-license")
    utime = (datetime.utcnow() - timedelta(hours=12)).timestamp()
    os.utime(cache, (utime, utime))

    assert local_run.get_stata_license(license_repo) == "repo-license"
    assert (tmp_path / "opensafely-stata.lic").read_text() == "repo-license"


def test_get_stata_license_repo_fetch(systmpdir, tmp_path, license_repo):
    assert local_run.get_stata_license(license_repo) == "repo-license"
    assert (tmp_path / "opensafely-stata.lic").read_text() == "repo-license"


def test_get_stata_license_repo_error(systmpdir):
    assert local_run.get_stata_license("/invalid/repo") is None


def test_add_arguments():
    parser = argparse.ArgumentParser(description="test")
    parser = local_run.add_arguments(parser)
    args = parser.parse_args(
        [
            "action",
            "--force-run-dependencies",
            "--project-dir=dir",
            "--continue-on-error",
            "--timestamps",
            "--debug",
        ]
    )

    assert args.actions == ["action"]
    assert args.force_run_dependencies
    assert args.project_dir == "dir"
    assert args.timestamps
    assert args.debug


def test_filter_log_messages():
    record = logging.makeLogRecord({})
    assert local_run.filter_log_messages(record)

    record = logging.makeLogRecord({"status_code": "code"})
    assert local_run.filter_log_messages(record)

    record = logging.makeLogRecord(
        {"status_code": local_run.StatusCode.WAITING_ON_DEPENDENCIES}
    )
    assert local_run.filter_log_messages(record) is False



"""
Development utility for creating and submitting a JobRequest without having a
job-server
"""
import argparse
import dataclasses
import pprint
import textwrap
from pathlib import Path
from urllib.parse import urlparse

from jobrunner.create_or_update_jobs import create_or_update_jobs
from jobrunner.lib.database import find_where
from jobrunner.lib.git import get_sha_from_remote_ref
from jobrunner.lib.log_utils import configure_logging
from jobrunner.models import Job, random_id
from jobrunner.sync import job_request_from_remote_format


def main(
    repo_url, actions, commit, branch, workspace, database, force_run_dependencies
):
    # Make paths to local repos absolute
    parsed = urlparse(repo_url)
    if not parsed.scheme and not parsed.netloc:
        path = Path(parsed.path).resolve()
        # In case we're on Windows
        repo_url = str(path).replace("\\", "/")
    if not commit:
        commit = get_sha_from_remote_ref(repo_url, branch)
    job_request = job_request_from_remote_format(
        dict(
            identifier=random_id(),
            sha=commit,
            workspace=dict(name=workspace, repo=repo_url, branch=branch, db=database),
            requested_actions=actions,
            force_run_dependencies=force_run_dependencies,
            cancelled_actions=[],
        )
    )
    print("Submitting JobRequest:\n")
    display_obj(job_request)
    create_or_update_jobs(job_request)
    jobs = find_where(Job, job_request_id=job_request.id)
    print(f"Created {len(jobs)} new jobs:\n")
    for job in jobs:
        display_obj(job)


def display_obj(obj):
    if hasattr(obj, "asdict"):
        data = obj.asdict()
    else:
        data = dataclasses.asdict(obj)
    output = pprint.pformat(data)
    print(textwrap.indent(output, "  "))
    print()


def run():
    configure_logging()
    parser = argparse.ArgumentParser(description=__doc__.partition("\n\n")[0])
    parser.add_argument("repo_url", help="URL (or local path) of git repository")
    parser.add_argument("actions", nargs="+", help="Name of project action to run")
    parser.add_argument(
        "--commit",
        help=(
            "Git commit to use (if repo_url is a local checkout, use current "
            "checked out commit by default)"
        ),
    )
    parser.add_argument(
        "--branch",
        help="Git branch or ref to use if no commit supplied (default HEAD)",
        default="HEAD",
    )
    parser.add_argument(
        "--workspace", help="Workspace ID (default 'test')", default="test"
    )
    parser.add_argument(
        "--database", help="Database name (default 'dummy')", default="dummy"
    )
    parser.add_argument("-f", "--force-run-dependencies", action="store_true")

    args = parser.parse_args()
    main(**vars(args))


if __name__ == "__main__":
    run()



"""
Defines the basic data structures which we pass around and store in the
database.

The `database` module contains some very basic code for storing/retrieving
these objects.

Note the schema is defined separately in `schema.sql`.
"""
import base64
import dataclasses
import datetime
import hashlib
import secrets
from enum import Enum

from jobrunner.lib.string_utils import project_name_from_url, slugify


class State(Enum):
    PENDING = "pending"
    RUNNING = "running"
    FAILED = "failed"
    SUCCEEDED = "succeeded"


# In contrast to State, these play no role in the state machine controlling
# what happens with a job. They are simply machine readable versions of the
# human readable status_message which allow us to provide certain UX
# affordances in the web and command line interfaces. These get added as we
# have a direct need for them, hence the minimal list below.
class StatusCode(Enum):
    WAITING_ON_DEPENDENCIES = "waiting_on_dependencies"
    DEPENDENCY_FAILED = "dependency_failed"
    WAITING_ON_WORKERS = "waiting_on_workers"
    NONZERO_EXIT = "nonzero_exit"
    CANCELLED_BY_USER = "cancelled_by_user"


# This is our internal representation of a JobRequest which we pass around but
# never save to the database (hence no __tablename__ attribute)
@dataclasses.dataclass
class JobRequest:
    id: str  # noqa: A003
    repo_url: str
    commit: str
    requested_actions: list
    cancelled_actions: list
    workspace: str
    database_name: str
    force_run_dependencies: bool = False
    force_run_failed: bool = False
    branch: str = None
    original: dict = None


# This stores the original JobRequest as received from the job-server. Once
# we've created the relevant Jobs we have no real need for the JobRequest
# object, but we it's useful to store it for debugging/audit purposes so we
# just save a blob of the original JSON as received from the job-server.
@dataclasses.dataclass
class SavedJobRequest:
    __tablename__ = "job_request"

    id: str  # noqa: A003
    original: dict


@dataclasses.dataclass
class Job:
    __tablename__ = "job"

    id: str = None  # noqa: A003
    job_request_id: str = None
    state: State = None
    # Git repository URL
    repo_url: str = None
    # Full commit sha
    commit: str = None
    # Name of workspace (effectively, the output directory)
    workspace: str = None
    # Only applicable to "generate_cohort" jobs: the name of the database to
    # query against
    database_name: str = None
    # Name of the action (one of the keys in the `actions` dict in
    # project.yaml)
    action: str = None
    # URL of git repository for action (None if action is not reusable)
    action_repo_url: str = None
    # Full SHA of commit in action repo (None if action is not reusable)
    action_commit: str = None
    # List of action names whose outputs need to be used as inputs to this
    # action
    requires_outputs_from: list = None
    # List of job IDs we need to wait to finish before we can run (these will
    # represent the subset of the actions above which hadn't already run when
    # this job was scheduled)
    wait_for_job_ids: list = None
    # The docker run arguments to execute
    run_command: str = None
    # The specific docker image that was actually run
    image_id: str = None
    # The specification of what outputs this job expects to produce, as a bunch
    # of named glob patterns organised by privacy level
    output_spec: dict = None
    # The outputs the job did produce matching the patterns above, as a mapping
    # of filenames to privacy levels
    outputs: dict = None
    # A list of the outputs the job produced which didn't match any of the
    # output patterns. This is only populated in the case that there are
    # unmatched output patterns, and is only used for debugging purposes.
    unmatched_outputs: list = None
    # Human readable string giving details about what's currently happening
    # with this job
    status_message: str = None
    # Machine readable code representing the status_message above
    status_code: StatusCode = None
    # Flag indicating that the user has cancelled this job
    cancelled: bool = False
    # Times (stored as integer UNIX timestamps)
    created_at: int = None
    updated_at: int = None
    started_at: int = None
    completed_at: int = None

    def __post_init__(self):
        # Generate a Job ID based on the Job Request ID and action. This means
        # we will always generate the same set of job IDs from a given Job
        # Request and so we won't create "orphan" jobs if we have to recreate
        # the job-runner database mid-job.
        #
        # Actions must be unique within a Job Request so this pair is
        # sufficient to give us global uniqueness. In fact we could do away
        # with Job IDs altogether and just use the action name directly, but
        # doing things this way is a less invasive change.
        if not self.id and self.job_request_id and self.action:
            self.id = deterministic_id(f"{self.job_request_id}\n{self.action}")

    def asdict(self):
        data = dataclasses.asdict(self)
        for key, value in data.items():
            # Convert Enums to strings for straightforward JSON serialization
            if isinstance(value, Enum):
                data[key] = value.value
            # Convert UNIX timestamp to ISO format
            elif isinstance(value, int) and key.endswith("_at"):
                data[key] = timestamp_to_isoformat(value)
        return data

    @property
    def created_at_isoformat(self):
        return timestamp_to_isoformat(self.created_at)

    @property
    def updated_at_isoformat(self):
        return timestamp_to_isoformat(self.updated_at)

    @property
    def started_at_isoformat(self):
        return timestamp_to_isoformat(self.started_at)

    @property
    def completed_at_isoformat(self):
        return timestamp_to_isoformat(self.completed_at)

    # On Python 3.8 we could use `functools.cached_property` here and avoid
    # recomputing this every time
    @property
    def project(self):
        """Project name based on github url."""
        return project_name_from_url(self.repo_url)

    # On Python 3.8 we could use `functools.cached_property` here and avoid
    # recomputing this every time
    @property
    def slug(self):
        """
        Use a human-readable slug rather than just an opaque ID to identify jobs in
        order to make debugging easier
        """
        return slugify(f"{self.project}-{self.action}-{self.id}")

    @property
    def output_files(self):
        if self.outputs:
            return self.outputs.keys()
        else:
            return []


def deterministic_id(seed):
    digest = hashlib.sha1(seed.encode("utf-8")).digest()
    return base64.b32encode(digest[:10]).decode("ascii").lower()


def random_id():
    return secrets.token_hex(5)


def timestamp_to_isoformat(ts):
    if ts is None:
        return None
    return datetime.datetime.utcfromtimestamp(ts).isoformat() + "Z"


def isoformat_to_timestamp(string):
    return int(
        datetime.datetime.fromisoformat(string.rstrip("Z") + "+00:00")
        .astimezone(datetime.timezone.utc)
        .timestamp()
    )



import os

from setuptools import find_packages, setup


with open(os.path.join("VERSION")) as f:
    version = f.read().strip()

setup(
    name="opensafely-jobrunner",
    version=version,
    packages=find_packages(exclude=["tests"]),
    include_package_data=True,
    url="https://github.com/opensafely-core/job-runner",
    author="OpenSAFELY",
    author_email="tech@opensafely.org",
    python_requires=">=3.8",
    install_requires=["ruamel.yaml", "requests"],
    classifiers=["License :: OSI Approved :: GNU General Public License v3 (GPLv3)"],
    entry_points=dict(
        console_scripts=[
            "local_run=jobrunner.cli.local_run:run",
            "add_job=jobrunner.cli.add_job:run",
            "kill_job=jobrunner.cli.kill_job:run",
            "retry_job=jobrunner.cli.retry_job:run",
        ],
    ),
)



import logging
import time
from datetime import datetime

from jobrunner.cli import local_run
from jobrunner.lib import log_utils
from jobrunner.models import Job, JobRequest


FROZEN_TIMESTAMP = 1608568119.1467905
FROZEN_TIMESTRING = datetime.utcfromtimestamp(FROZEN_TIMESTAMP).isoformat()

repo_url = "https://github.com/opensafely/project"
test_job = Job(id="id", action="action", repo_url=repo_url)
test_request = JobRequest(
    id="request",
    repo_url=repo_url,
    workspace="workspace",
    commit="commit",
    requested_actions=["action"],
    cancelled_actions=[],
    database_name="dummy",
)


def test_formatting_filter():
    record = logging.makeLogRecord({})
    assert log_utils.formatting_filter(record)
    assert record.action == ""

    record = logging.makeLogRecord({"job": test_job})
    assert log_utils.formatting_filter(record)
    assert record.action == "action: "
    assert record.tags == "project=project action=action id=id"

    record = logging.makeLogRecord({"job": test_job, "status_code": "code"})
    assert log_utils.formatting_filter(record)
    assert record.tags == "status=code project=project action=action id=id"

    test_job2 = Job(id="id", action="action", repo_url=repo_url, status_code="code")
    record = logging.makeLogRecord({"job": test_job2})
    assert log_utils.formatting_filter(record)
    assert record.tags == "status=code project=project action=action id=id"

    record = logging.makeLogRecord({"job": test_job, "job_request": test_request})
    assert log_utils.formatting_filter(record)
    assert record.tags == "project=project action=action id=id req=request"

    record = logging.makeLogRecord({"status_code": ""})
    assert log_utils.formatting_filter(record)
    assert record.tags == ""


def test_formatting_filter_with_context():
    record = logging.makeLogRecord({})
    with log_utils.set_log_context(job=test_job):
        assert log_utils.formatting_filter(record)
    assert record.action == "action: "
    assert record.tags == "project=project action=action id=id"

    record = logging.makeLogRecord({"status_code": "code"})
    with log_utils.set_log_context(job=test_job):
        assert log_utils.formatting_filter(record)
    assert record.tags == "status=code project=project action=action id=id"

    record = logging.makeLogRecord({})
    with log_utils.set_log_context(job=test_job, job_request=test_request):
        assert log_utils.formatting_filter(record)
    assert record.tags == "project=project action=action id=id req=request"


def test_jobrunner_formatter_default(monkeypatch):
    monkeypatch.setattr(time, "time", lambda: FROZEN_TIMESTAMP)
    record = logging.makeLogRecord(
        {
            "msg": "message",
            "job": test_job,
            "job_request": test_request,
            "status_code": "status",
        }
    )
    log_utils.formatting_filter(record)
    formatter = log_utils.JobRunnerFormatter(log_utils.DEFAULT_FORMAT, style="{")
    assert formatter.format(record) == (
        "2020-12-21 16:28:39.146Z message "
        "status=status project=project action=action id=id req=request"
    )


def test_jobrunner_formatter_local_run(monkeypatch):
    monkeypatch.setattr(time, "time", lambda: FROZEN_TIMESTAMP)
    record = logging.makeLogRecord(
        {
            "msg": "message",
            "job": test_job,
            "job_request": test_request,
            "status_code": "status",
        }
    )
    log_utils.formatting_filter(record)
    formatter = log_utils.JobRunnerFormatter(local_run.LOCAL_RUN_FORMAT, style="{")
    assert formatter.format(record) == "action: message"



"""
Quick script to grab an extract of resource usage stats from the stats database
"""
import argparse
import gzip
import sqlite3
from pathlib import Path

from jobrunner import config


def main(output_file, since, gz=False):
    output_file = Path(output_file)
    database_file = config.STATS_DATABASE_FILE
    assert database_file.exists()
    assert not output_file.exists()
    conn = sqlite3.connect(database_file)
    conn.execute("ATTACH DATABASE ? AS extract", [str(output_file)])
    conn.execute(
        """
        CREATE TABLE extract.containers AS SELECT
          timestamp,
          json_each.key AS container,
          json_extract(json_each.value, "$.cpu_percentage") AS cpu_percentage,
          json_extract(json_each.value, "$.memory_used") AS memory_used
        FROM
          main.stats,
          json_each(data, '$.containers')
        WHERE
           timestamp > ?
        """,
        [since],
    )
    conn.close()
    if gz:
        output_file.with_suffix(".sqlite.gz").write_bytes(
            gzip.compress(output_file.read_bytes(), compresslevel=6)
        )
        output_file.unlink()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description=__doc__.partition("\n\n")[0])
    parser.add_argument("output_file", help="File to write to")
    parser.add_argument(
        "--since", help="Extract stats since this date", default="2020-01-01"
    )
    parser.add_argument("--gz", action="store_true", help="Gzip output file")
    args = parser.parse_args()
    main(**vars(args))



"""
Super crude docker/system stats logger
"""
import datetime
import json
import logging
import sqlite3
import sys
import time

from jobrunner import config
from jobrunner.lib.docker_stats import (
    get_container_stats,
    get_volume_and_container_sizes,
)
from jobrunner.lib.log_utils import configure_logging
from jobrunner.lib.system_stats import DockerDiskSpaceError, get_system_stats


SCHEMA_SQL = """
CREATE TABLE stats (
    timestamp TEXT,
    data TEXT
);
"""


log = logging.getLogger(__name__)


def main():
    database_file = config.STATS_DATABASE_FILE
    if not database_file:
        log.info("STATS_DATABASE_FILE not set; not polling for system stats")
        return
    log.info(f"Logging system stats to: {database_file}")
    connection = get_database_connection(database_file)
    while True:
        log_stats(connection)
        time.sleep(config.STATS_POLL_INTERVAL)


def get_database_connection(filename):
    filename.parent.mkdir(parents=True, exist_ok=True)
    conn = sqlite3.connect(filename)
    # Enable autocommit
    conn.isolation_level = None
    schema_count = list(conn.execute("SELECT COUNT(*) FROM sqlite_master"))[0][0]
    if schema_count == 0:
        conn.executescript(SCHEMA_SQL)
    return conn


def log_stats(connection):
    stats = get_all_stats()
    # If no containers are running then don't log anything
    if not stats["containers"]:
        return
    timestamp = datetime.datetime.utcnow().isoformat()
    connection.execute(
        "INSERT INTO stats (timestamp, data) VALUES (?, ?)",
        [timestamp, json.dumps(stats)],
    )


def get_all_stats():
    try:
        stats = get_system_stats()
    except DockerDiskSpaceError:
        # Sometimes we're so low on disk space that we can't start the
        # container needed to get system-level stats, in this case it's better
        # to not get the stats than to bail entirely (and spam the logs in the
        # process)
        stats = {}
    volume_sizes, container_sizes = get_volume_and_container_sizes()
    containers = get_container_stats()
    for name, container in containers.items():
        container["disk_used"] = container_sizes.get(name)
    stats["containers"] = containers
    stats["volumes"] = volume_sizes
    return stats


if __name__ == "__main__":
    configure_logging()

    try:
        main()
    except KeyboardInterrupt:
        sys.exit(0)



import dataclasses
import json
import posixpath
import shlex
from pathlib import PurePosixPath, PureWindowsPath
from types import SimpleNamespace

from jobrunner import config
from jobrunner.lib.yaml_utils import YAMLError, parse_yaml


# The magic action name which means "run every action"
RUN_ALL_COMMAND = "run_all"

# The version of `project.yaml` where each feature was introduced
FEATURE_FLAGS_BY_VERSION = {"UNIQUE_OUTPUT_PATH": 2, "EXPECTATIONS_POPULATION": 3}


class ProjectValidationError(Exception):
    pass


class UnknownActionError(ProjectValidationError):
    pass


class InvalidPatternError(ProjectValidationError):
    pass


# Tiny dataclass to capture the specification of a project action
@dataclasses.dataclass
class ActionSpecifiction:
    run: str
    needs: list
    outputs: dict


def parse_and_validate_project_file(project_file):
    """Parse and validate the project file.

    Args:
        project_file: The contents of the project file as an immutable array of bytes.

    Returns:
        A dict representing the project.

    Raises:
        ProjectValidationError: The project could not be parsed, or was not valid
    """
    try:
        project = parse_yaml(project_file, name="project.yaml")
    except YAMLError as e:
        raise ProjectValidationError(*e.args)
    return validate_project_and_set_defaults(project)


# Copied almost verbatim from the original job-runner
def validate_project_and_set_defaults(project):
    """Check that a dictionary of project actions is valid, and set any defaults"""
    feat = get_feature_flags_for_version(project.get("version"))
    seen_runs = []
    seen_output_files = []
    if feat.EXPECTATIONS_POPULATION:
        if "expectations" not in project:
            raise ProjectValidationError("Project must include `expectations` section")
        if "population_size" not in project["expectations"]:
            raise ProjectValidationError(
                "Project `expectations` section must include `population` section",
            )
        try:
            int(project["expectations"]["population_size"])
        except TypeError:
            raise ProjectValidationError(
                "Project expectations population size must be a number",
            )
    else:
        project["expectations"] = {}
        project["expectations"]["population_size"] = 1000

    project_actions = project["actions"]

    for action_id, action_config in project_actions.items():
        if is_generate_cohort_command(shlex.split(action_config["run"])):
            if len(action_config["outputs"]) != 1:
                raise ProjectValidationError(
                    f"A `generate_cohort` action must have exactly one output; {action_id} had {len(action_config['outputs'])}",
                )

        # Check a `generate_cohort` command only generates a single output
        # Check outputs are permitted
        for privacy_level, output in action_config["outputs"].items():
            permitted_privacy_levels = [
                "highly_sensitive",
                "moderately_sensitive",
                "minimally_sensitive",
            ]
            if privacy_level not in permitted_privacy_levels:
                raise ProjectValidationError(
                    f"{privacy_level} is not valid (must be one of {', '.join(permitted_privacy_levels)})",
                )

            for output_id, filename in output.items():
                try:
                    assert_valid_glob_pattern(filename)
                except InvalidPatternError as e:
                    raise ProjectValidationError(
                        f"Output path {filename} is not permitted: {e}"
                    )

                if feat.UNIQUE_OUTPUT_PATH and filename in seen_output_files:
                    raise ProjectValidationError(
                        f"Output path {filename} is not unique"
                    )
                seen_output_files.append(filename)

        command, *args = shlex.split(action_config["run"])
        name, _, version = command.partition(":")
        if not version:
            raise ProjectValidationError(
                f"{name} must have a version specified (e.g. {name}:0.5.2)",
            )
        # Check the run command + args signature appears only once in
        # a project
        run_signature = f"{name}_{args}"
        if run_signature in seen_runs:
            raise ProjectValidationError(
                f"{name} {' '.join(args)} appears more than once"
            )
        seen_runs.append(run_signature)

        for dependency in action_config.get("needs", []):
            if dependency not in project_actions:
                if " " in dependency:
                    raise ProjectValidationError(
                        f"`needs` actions in '{action_id}' should be separated"
                        f" with commas:\n{', '.join(dependency.split())}"
                    )
                raise ProjectValidationError(
                    f"Action '{action_id}' lists unknown action '{dependency}'"
                    f" in its `needs` config"
                )

    return project


def get_action_specification(project, action_id):
    """Get a specification for the action from the project.

    Args:
        project: A dict representing the project.
        action_id: The string ID of the action.

    Returns:
        An instance of ActionSpecification.

    Raises:
        UnknownActionError: The action was not found in the project.
        ProjectValidationError: The project was not valid.
    """
    try:
        action_spec = project["actions"][action_id]
    except KeyError:
        raise UnknownActionError(f"Action '{action_id}' not found in project.yaml")
    run_command = action_spec["run"]
    if "config" in action_spec:
        run_command = add_config_to_run_command(run_command, action_spec["config"])
    run_args = shlex.split(run_command)

    # Special case handling for the `cohortextractor generate_cohort` command
    if is_generate_cohort_command(run_args, require_version=1):
        # Set the size of the dummy data population, if that's what we're
        # generating.  Possibly this should be moved to the study definition
        # anyway, which would make this unnecessary.
        if config.USING_DUMMY_DATA_BACKEND:
            if "dummy_data_file" in action_spec:
                run_command += f" --dummy-data-file={action_spec['dummy_data_file']}"
            else:
                size = int(project["expectations"]["population_size"])
                run_command += f" --expectations-population={size}"
        # Automatically configure the cohortextractor to produce output in the
        # directory the `outputs` spec is expecting. Longer term I'd like to
        # just make it an error if the directories don't match, rather than
        # silently fixing it. (We can use the project versioning system to
        # ensure this doesn't break existing studies.)
        output_dirs = get_output_dirs(action_spec["outputs"])
        if len(output_dirs) != 1:
            # If we detect multiple output directories but the command
            # explicitly specifies an output directory then we assume the user
            # knows what they're doing and don't attempt to modify the output
            # directory or throw an error
            if not args_include(run_args, "--output-dir"):
                raise ProjectValidationError(
                    f"generate_cohort command should produce output in only one "
                    f"directory, found {len(output_dirs)}:\n"
                    + "\n".join([f" - {d}/" for d in output_dirs])
                )
        else:
            run_command += f" --output-dir={output_dirs[0]}"

    elif is_generate_cohort_command(run_args, require_version=2):
        # cohortextractor Version 2 expects all command line arguments to be
        # specified in the run command
        if config.USING_DUMMY_DATA_BACKEND and "--dummy-data-file" not in run_command:
            raise ProjectValidationError(
                "--dummy-data-file is required for a local run"
            )

        # There is one and only one output file in the outputs spec (verified
        # in validate_project_and_set_defaults())
        output_file = next(
            output_file
            for output in action_spec["outputs"].values()
            for output_file in output.values()
        )
        if output_file not in run_command:
            raise ProjectValidationError(
                "--output in run command and outputs must match"
            )

    elif is_generate_cohort_command(run_args):
        raise RuntimeError("Unhandled cohortextractor version")

    return ActionSpecifiction(
        run=run_command,
        needs=action_spec.get("needs", []),
        outputs=action_spec["outputs"],
    )


def add_config_to_run_command(run_command, config):
    """Add --config flag to command.

    For commands that require complex config, users can supply a config key in
    project.yaml.  We serialize this as JSON, and pass it to the command with the
    --config flag.
    """
    config_as_json = json.dumps(config).replace("'", r"\u0027")
    return f"{run_command} --config '{config_as_json}'"


def requires_db_access(args):
    """
    By default actions do not have database access, but certain trusted actions require it
    """
    valid_commands = {
        "cohortextractor": ("generate_cohort", "generate_codelist_report"),
        "cohortextractor-v2": ("generate_cohort", "generate_dataset"),
        "databuilder": ("generate_dataset",),
    }
    if len(args) <= 1:
        return False

    image, command = args[0], args[1]
    image = image.split(":")[0]
    return command in valid_commands.get(image, [])


def is_generate_cohort_command(args, require_version=None):
    """
    The `cohortextractor generate_cohort` command gets special treatment in
    various places (e.g. it's the only command which gets access to the
    database) so it's helpful to have a single function for identifying it
    """
    assert not isinstance(args, str)
    version_found = None
    if len(args) > 1 and args[1] in ("generate_cohort", "generate_dataset"):
        if args[0].startswith("cohortextractor:"):
            version_found = 1
        # databuilder is a rebranded cohortextractor-v2.
        # Retain cohortextractor-v2 for backwards compatibility for now.
        elif args[0].startswith(("cohortextractor-v2:", "databuilder:")):
            version_found = 2
    # If we're not looking for a specific version then return True if any
    # version found
    if require_version is None:
        return version_found is not None
    # Otherwise return True only if specified version found
    else:
        return version_found == require_version


def args_include(args, target_arg):
    return any(arg == target_arg or arg.startswith(f"{target_arg}=") for arg in args)


def get_all_actions(project):
    # We ignore any manually defined run_all action (in later project versions
    # this will be an error). We use a list comprehension rather than set
    # operators as previously so we preserve the original order.
    return [action for action in project["actions"].keys() if action != RUN_ALL_COMMAND]


def get_all_output_patterns_from_project_file(project_file):
    project = parse_and_validate_project_file(project_file)
    all_patterns = set()
    for action in project["actions"].values():
        for patterns in action["outputs"].values():
            all_patterns.update(patterns.values())
    return list(all_patterns)


def get_output_dirs(output_spec):
    """
    Given the set of output files specified by an action, return a list of the
    unique directory names of those outputs
    """
    filenames = []
    for group in output_spec.values():
        filenames.extend(group.values())
    dirs = set(PurePosixPath(filename).parent for filename in filenames)
    return list(dirs)


def get_feature_flags_for_version(version):
    latest_version = max(FEATURE_FLAGS_BY_VERSION.values())
    if version is None:
        raise ProjectValidationError(
            f"Project file must have a `version` attribute specifying which "
            f"version of the project configuration format it uses (current "
            f"latest version is {latest_version})"
        )
    try:
        version = float(version)
    except (TypeError, ValueError):
        raise ProjectValidationError(
            f"`version` must be a number between 1 and {latest_version}"
        )
    feat = SimpleNamespace()
    matched_any = False
    for k, v in FEATURE_FLAGS_BY_VERSION.items():
        if v <= version:
            setattr(feat, k, True)
            matched_any = True
        else:
            setattr(feat, k, False)
    if version > 1 and not matched_any:
        raise ProjectValidationError(
            f"`version` must be a number between 1 and {latest_version}"
        )
    return feat


def assert_valid_glob_pattern(pattern):
    """
    These patterns get converted into regular expressions and matched
    with a `find` command so there shouldn't be any possibility of a path
    traversal attack anyway. But it's still good to ensure that they are
    well-formed.
    """
    # Only POSIX slashes please
    if "\\" in pattern:
        raise InvalidPatternError("contains back slashes (use forward slashes only)")
    # These aren't unsafe, but they won't behave as expected so we shouldn't let
    # people use them
    for expr in ("**", "?", "["):
        if expr in pattern:
            raise InvalidPatternError(
                f"contains '{expr}' (only the * wildcard character is supported)"
            )
    if pattern.endswith("/"):
        raise InvalidPatternError(
            "looks like a directory (only files should be specified)"
        )
    # Check that the path is in normal form
    if posixpath.normpath(pattern) != pattern:
        raise InvalidPatternError(
            "is not in standard form (contains double slashes or '..' elements)"
        )
    # This is the directory we use for storing metadata about action runs and
    # we don't want outputs getting mixed up in it.
    if pattern == "metadata" or pattern.startswith("metadata/"):
        raise InvalidPatternError("should not include the metadata directory")
    # Windows has a different notion of absolute paths (e.g c:/foo) so we check
    # for both platforms
    if PurePosixPath(pattern).is_absolute() or PureWindowsPath(pattern).is_absolute():
        raise InvalidPatternError("is an absolute path")



"""
Utility functions for interacting with Docker
"""
import json
import os
import re
import subprocess

from jobrunner import config
from jobrunner.lib import atomic_writer
from jobrunner.lib.subprocess_utils import subprocess_run


# Docker requires a container in order to interact with volumes, but it doesn't
# much matter what it is for our purposes as long as it has `sh` and `find`
MANAGEMENT_CONTAINER_IMAGE = f"{config.DOCKER_REGISTRY}/busybox"

# This path is pretty arbitrary: it sets where we mount volumes inside their
# management containers (which are used for copying files in and out), but this
# is independent of where the volumes get mounted inside other containers to
# which they may get attached
VOLUME_MOUNT_POINT = "/workspace"

# Apply this label (Docker-speak for "tag") to all containers and volumes we
# create for easier management and test cleanup
LABEL = "job-runner"


# Shelling out to the Docker client on Windows sometimes hangs indefinitely
# (for no obvious reason that we can determine). To prevent this locking up the
# entire process we use a large-ish default timeout on almost all calls to
# Docker. (Timeout value is in seconds.)
DEFAULT_TIMEOUT = 5 * 60


class DockerPullError(Exception):
    pass


class DockerAuthError(DockerPullError):
    pass


class DockerTimeoutError(Exception):
    pass


class DockerDiskSpaceError(Exception):
    pass


def add_docker_labels(cmd, labels):
    """Add labels to a docker cmd."""
    if not labels:
        return
    for name, value in labels.items():
        cmd.extend(["--label", f"{name}={value}"])


def docker(docker_args, timeout=DEFAULT_TIMEOUT, **kwargs):
    args = ["docker"] + docker_args
    try:
        return subprocess_run(args, timeout=timeout, **kwargs)
    except subprocess.TimeoutExpired as e:
        raise DockerTimeoutError from e
    except subprocess.CalledProcessError as e:
        output = e.stderr
        if output is None:
            output = e.stdout
        if isinstance(output, bytes):
            output = output.decode("utf8", "ignore")
        if (
            output is not None
            and e.returncode == 1
            and "Error response from daemon:" in output
            and ": no space left on device" in output
        ):
            raise DockerDiskSpaceError from e
        else:
            raise


def create_volume(volume_name, labels=None):
    """
    Creates the named volume and also creates (but does not start) a "manager"
    container which we can use to copy files in and out of the volume. Note
    that in order to interact with the volume a container with that volume
    mounted must exist, but it doesn't need to be running.
    """
    cmd = ["volume", "create", "--label", LABEL, "--name", volume_name]
    add_docker_labels(cmd, labels)
    docker(cmd, check=True, capture_output=True)
    # Run a basic container that mounts this image.  Having the volume mounted
    # allows us to copy from/to it, and having the container running protects
    # it from rogue `docker container prune` commands.
    try:
        run(
            manager_name(volume_name),
            [MANAGEMENT_CONTAINER_IMAGE, "sh"],
            volume=(volume_name, VOLUME_MOUNT_POINT),
            label=LABEL,
            labels=labels,
            extra_args=[
                "--interactive",
                "--restart=unless-stopped",
            ],
        )
    except subprocess.CalledProcessError as e:
        # If a volume and its manager already exist we don't want to throw an
        # error. `docker volume create` is naturally idempotent, but we have to
        # handle this manually here.
        if e.returncode != 125 or b"is already in use by container" not in e.stderr:
            raise


def volume_exists(volume_name):
    """Does the given volume exist?"""
    try:
        docker(["volume", "inspect", volume_name], check=True, capture_output=True)
    except subprocess.CalledProcessError:
        return False
    else:
        return True


def delete_volume(volume_name):
    """
    Deletes the named volume and its manager container
    """
    try:
        docker(
            ["rm", "--force", manager_name(volume_name)],
            check=True,
            capture_output=True,
        )
    except subprocess.CalledProcessError as e:
        # Ignore error if container has already been removed
        if e.returncode != 1 or b"No such container" not in e.stderr:
            raise
    try:
        docker(
            [
                "volume",
                "rm",
                volume_name,
            ],
            check=True,
            capture_output=True,
        )
    except subprocess.CalledProcessError as e:
        # Ignore error if container has already been removed
        if e.returncode != 1 or b"No such volume" not in e.stderr:
            raise


def copy_to_volume(volume_name, source, dest, timeout=None):
    """
    Copy the contents of `directory` to the root of the named volume

    As this command can potentially take a long time with large files it does
    not, by default, have any timeout.
    """
    if source.is_dir():
        # Ensure the *contents* of the directory are copied, rather than the
        # directory itself. See:
        # https://docs.docker.com/engine/reference/commandline/cp/#extended-description
        source = str(source).rstrip("/") + "/."
    docker(
        [
            "cp",
            "--follow-link",
            source,
            f"{manager_name(volume_name)}:{VOLUME_MOUNT_POINT}/{dest}",
        ],
        check=True,
        capture_output=True,
        timeout=timeout,
    )


def copy_from_volume(volume_name, source, dest, timeout=None):
    """
    Copy the contents of `source` from the root of the named volume to `dest`
    on local disk

    As this command can potentially take a long time with large files it does
    not, by default, have any timeout.
    """
    with atomic_writer(dest) as tmp:
        docker(
            [
                "cp",
                "--follow-link",
                f"{manager_name(volume_name)}:{VOLUME_MOUNT_POINT}/{source}",
                tmp,
            ],
            check=True,
            capture_output=True,
            timeout=timeout,
        )


def glob_volume_files(volume_name, glob_patterns):
    """
    Accept a list of glob patterns and return a dict mapping each pattern to a
    list of all the files in `volume_name` which match

    Accepting multiple patterns like this allow us to avoid multiple round
    trips through Docker when we need to match several different patterns.
    """
    # Guard against the easy mistake of passing a single string pattern, rather
    # than a list of patterns
    assert not isinstance(glob_patterns, str)
    # Build a `find` command
    args = ["find", VOLUME_MOUNT_POINT, "-type", "f", "("]
    # We need to use regex matching rather than `-path` because find's
    # wildcards are too liberal and match across path separators (e.g
    # "foo/*.py" matches Python files in all sub-directories of "foo" rather
    # than just the top level)
    for pattern in glob_patterns:
        args.extend(
            ["-regex", _glob_pattern_to_regex(f"{VOLUME_MOUNT_POINT}/{pattern}"), "-o"]
        )
    # Replace final OR flag with a closing bracket
    args[-1] = ")"
    response = docker(
        ["container", "exec", manager_name(volume_name)] + args,
        check=True,
        capture_output=True,
        text=True,
        encoding="utf-8",
    )
    # Remove the volume path prefix from the results
    chars_to_strip = len(VOLUME_MOUNT_POINT) + 1
    files = [f[chars_to_strip:] for f in response.stdout.splitlines()]
    files = sorted(files)
    matches = {}
    for pattern in glob_patterns:
        regex = re.compile(_glob_pattern_to_regex(pattern))
        matches[pattern] = [f for f in files if regex.match(f)]
    return matches


def _glob_pattern_to_regex(glob_pattern):
    """
    Convert a shell glob pattern (where the wildcard does not match the "/"
    character) into a regular expression
    """
    literals = glob_pattern.split("*")
    return "[^/]*".join(map(re.escape, literals))


def find_newer_files(volume_name, reference_file):
    """
    Return all files in volume newer than the reference file
    """
    args = [
        "find",
        VOLUME_MOUNT_POINT,
        "-type",
        "f",
        "-newer",
        f"{VOLUME_MOUNT_POINT}/{reference_file}",
    ]
    response = docker(
        ["container", "exec", manager_name(volume_name)] + args,
        check=True,
        capture_output=True,
        text=True,
        encoding="utf-8",
    )
    # Remove the volume path prefix from the results
    chars_to_strip = len(VOLUME_MOUNT_POINT) + 1
    files = [f[chars_to_strip:] for f in response.stdout.splitlines()]
    return sorted(files)


def manager_name(volume_name):
    return f"{volume_name}-manager"


def container_exists(name):
    return bool(container_inspect(name, "ID", none_if_not_exists=True))


def container_is_running(name):
    return container_inspect(name, "State.Running", none_if_not_exists=True) or False


def container_inspect(name, key="", none_if_not_exists=False):
    """
    Retrieves metadata about the named container. By default will return
    everything but `key` can be a dotted path to a specific piece of metadata.

    Optionally returns None if the container does not exist

    See: https://docs.docker.com/engine/reference/commandline/inspect/
    """
    try:
        response = docker(
            ["container", "inspect", "--format", "{{json .%s}}" % key, name],
            check=True,
            capture_output=True,
        )
    except subprocess.CalledProcessError as e:
        if (
            none_if_not_exists
            and e.returncode == 1
            and b"No such container" in e.stderr
        ):
            return
        else:
            raise
    return json.loads(response.stdout)


def run(
    name,
    args,
    volume=None,
    env=None,
    allow_network_access=False,
    label=None,
    labels=None,
    extra_args=None,
):
    run_args = ["run", "--init", "--detach", "--label", LABEL, "--name", name]
    if extra_args is not None:
        run_args.extend(extra_args)

    if not allow_network_access:
        run_args.extend(["--network", "none"])
    if volume:
        run_args.extend(["--volume", f"{volume[0]}:{volume[1]}"])
    # These lables are in addition to the default LABEL which is always applied
    # Single unary label
    if label is not None:
        run_args.extend(["--label", label])
    if labels:
        add_docker_labels(run_args, labels)
    # To avoid leaking the values into the command line arguments we set them
    # in the evnironment and tell Docker to fetch them from there
    if env is None:
        env = {}
    for key, value in env.items():
        run_args.extend(["--env", key])
    docker(
        run_args + args, check=True, capture_output=True, env=dict(os.environ, **env)
    )


def image_exists_locally(image_name_and_version):
    try:
        docker(
            ["image", "inspect", "--format", "ok", image_name_and_version],
            check=True,
            capture_output=True,
        )
        return True
    except subprocess.CalledProcessError as e:
        if e.returncode == 1 and b"No such image" in e.stderr:
            return False
        raise


def delete_container(name):
    try:
        docker(
            ["container", "rm", "--force", name],
            check=True,
            capture_output=True,
        )
    except subprocess.CalledProcessError as e:
        # Ignore error if container has already been removed
        if e.returncode != 1 or b"No such container" not in e.stderr:
            raise


def kill(name):
    try:
        docker(
            ["container", "kill", name],
            check=True,
            capture_output=True,
        )
    except subprocess.CalledProcessError as e:
        # Ignore error if container has already been killed or removed
        if e.returncode != 1 or (
            b"No such container" not in e.stderr and b"is not running" not in e.stderr
        ):
            raise


def write_logs_to_file(container_name, filename):
    with open(filename, "wb") as f:
        docker(
            ["container", "logs", "--timestamps", container_name],
            check=True,
            stdout=f,
            stderr=subprocess.STDOUT,
        )


def pull(image, quiet=False):
    try:
        docker(
            ["pull", image, *(["--quiet"] if quiet else [])],
            check=True,
            encoding="utf-8",
            # When not running "quiet" we don't capture stdout so that progress
            # gets shown in the terminal
            stdout=subprocess.PIPE if quiet else None,
            stderr=subprocess.PIPE,
            timeout=None,
        )
    except subprocess.CalledProcessError as e:
        message = e.stderr.strip()
        if message.endswith(": unauthorized"):
            raise DockerAuthError(message)
        else:
            raise DockerPullError(message)
